/******************************************************************
 *****                                                        *****
 *****  Name: cs8900.c                                        *****
 *****  Ver.: 1.0                                             *****
 *****  Date: 07/05/2001                                      *****
 *****  Auth: Andreas Dannenberg                              *****
 *****        HTWK Leipzig                                    *****
 *****        university of applied sciences                  *****
 *****        Germany                                         *****
 *****  Func: ethernet packet-driver for use with LAN-        *****
 *****        controller CS8900 from Crystal/Cirrus Logic     *****
 *****                                                        *****
 *****  Keil: Module modified for use with Philips            *****
 *****        LPC2378 EMAC Ethernet controller                *****
 *****                                                        *****
 ******************************************************************/

/* Adapted from file originally written by Andreas Dannenberg.  Supplied with permission. */

#include "Include.h"

#ifdef WITH_IP

#include "emac.h"

DWORD CheckPHYAddr(void);

/* The semaphore used to wake the uIP task when data arives. */

static unsigned short *rptr;
static unsigned short *tptr;

volatile DWORD phyAddr;
volatile DWORD phyTyp;

volatile DWORD Duplex;
volatile DWORD Speed;

void write_PHY (int PhyReg, int Value);
void rx_descr_init (void);
void tx_descr_init (void);
unsigned short ReadFrame_EMAC(void);
void WriteFrame_EMAC(unsigned short Data);

// easyWEB internal function
// help function to swap the byte order of a WORD

static unsigned short SwapBytes(unsigned short Data)
{
  return (Data >> 8) | (Data << 8);
}

// Keil: function added to write PHY
void write_PHY (int PhyReg, int Value)
{
  unsigned int tout;

  MAC_MADR = phyAddr | PhyReg;
  MAC_MWTD = Value;

  /* Wait utill operation completed */
  tout = 0;
  for (tout = 0; tout < MII_WR_TOUT; tout++) {
    if ((MAC_MIND & MIND_BUSY) == 0) {
      break;
    }
  }
}


// Keil: function added to read PHY
unsigned short read_PHY (unsigned char PhyReg)
{
  unsigned int tout;

  MAC_MADR = phyAddr | PhyReg;
  MAC_MCMD = MCMD_READ;

  /* Wait until operation completed */
  tout = 0;
  for (tout = 0; tout < MII_RD_TOUT; tout++) {
    if ((MAC_MIND & MIND_BUSY) == 0) {
      break;
    }
  }
  MAC_MCMD = 0;
  return (MAC_MRDD);
}

/*****************************************************************************
** Function name:		CheckPHYAddr
**
** Descriptions:		Find the Address from PHY
**
** parameters:			none
** Returned value:		PHY_ADDR
**
*****************************************************************************/
DWORD CheckPHYAddr(void)
{
	DWORD i = 0;
	phyAddr = DP83848C_DEF_ADR & ~0x1800;
	while (i < 4)
	{
		if (read_PHY(0) != 0xFFFF)
			return phyAddr;
		i++;
		phyAddr |= (i << 11);
	}
	phyAddr = DP83848C_DEF_ADR;
	return 0;
}

// Keil: function added to initialize Rx Descriptors
void rx_descr_init (void)
{
  unsigned int i;

  for (i = 0; i < NUM_RX_FRAG; i++) {
    RX_DESC_PACKET(i)  = RX_BUF(i);						// (*(unsigned int *)(0x7FE00000 + 8*i)) = ((0x7FE00058) + 600*i)
    RX_DESC_CTRL(i)    = RCTRL_INT | (ETH_FRAG_SIZE-1);	// (*(unsigned int *)(0x7FE00004 + 8*i)) = 0x800005FF
    RX_STAT_INFO(i)    = 0;								// (*(unsigned int *)(0x7FE00020 + 8*i))
    RX_STAT_HASHCRC(i) = 0;								// (*(unsigned int *)(0x7FE00024 + 8*i))
  }

  /* Set EMAC Receive Descriptor Registers. */
  MAC_RXDESCRIPTOR    = RX_DESC_BASE;		// 0x7FE00000
  MAC_RXSTATUS        = RX_STAT_BASE;		// 0x7FE00020
  MAC_RXDESCRIPTORNUM = NUM_RX_FRAG-1;		// 3

  /* Rx Descriptors Point to 0 */
  MAC_RXCONSUMEINDEX  = 0;
}


// Keil: function added to initialize Tx Descriptors
void tx_descr_init (void) {
  unsigned int i;

  for (i = 0; i < NUM_TX_FRAG; i++) {
    TX_DESC_PACKET(i) = TX_BUF(i);
    TX_DESC_CTRL(i)   = 0;
    TX_STAT_INFO(i)   = 0;
  }

  /* Set EMAC Transmit Descriptor Registers. */
  MAC_TXDESCRIPTOR    = TX_DESC_BASE;
  MAC_TXSTATUS        = TX_STAT_BASE;
  MAC_TXDESCRIPTORNUM = NUM_TX_FRAG-1;

  /* Tx Descriptors Point to 0 */
  MAC_TXPRODUCEINDEX  = 0;
}


// configure port-pins for use with LAN-controller,
// reset it and send the configuration-sequence

DWORD Init_EMAC(void)
{

// Keil: function modified to access the EMAC
// Initializes the EMAC ethernet controller
  	volatile unsigned int regv,id1,id2,i;

	static DWORD iStep = 0;
	static DWORD iStepOld = 0;
	if (iStep != iStepOld)
	{
		iStepOld = iStep;
		//Seriell_Text("\r\nStep: ");
		//Seriell_Int(iStep);
	}
	switch (iStep)
	{
	case 0:
		/* Enable P1 Ethernet Pins. */
		PINSEL2 = (PINSEL2 &  0x0FC0FCF0) | 0x50150105;
		PINSEL3 = (PINSEL3 & ~0x0000000F) | 0x00000005;

		/* Power Up the EMAC controller. */
		PCONP |= 0x40000000;
		iStep = 1;
		return 0;
		break;
	case 1:
		MAC_MCFG = 0x8018;	/* host clock divided by 20, no suppress preamble, no scan increment */
		for ( i = 0; i < 0x40; i++ );
		MAC_MCFG = 0x0018;	/* Apply a reset */

		/* Reset all EMAC internal modules. */
		MAC_MAC1 = MAC1_RES_TX | MAC1_RES_MCS_TX | MAC1_RES_RX | MAC1_RES_MCS_RX | MAC1_SIM_RES | MAC1_SOFT_RES;
		MAC_COMMAND = CR_REG_RES | CR_TX_RES | CR_RX_RES;

		/* A short delay after reset. */
		iStep = 2;
		return 0;
		break;
	case 2:

		/* Initialize MAC control registers. */
		MAC_MAC1 = MAC1_PASS_ALL;
		MAC_MAC2 = MAC2_CRC_EN | MAC2_PAD_EN;
		MAC_MAXF = ETH_MAX_FLEN;
		MAC_CLRT = CLRT_DEF;
		MAC_IPGR = IPGR_DEF;

		/* Enable Reduced MII interface. */
		MAC_COMMAND = CR_RMII | CR_PASS_RUNT_FRM;

		/* Reset Reduced MII Logic. */
		MAC_SUPP = SUPP_RES_RMII;
		MAC_SUPP = 0;

		// Additional find out the PHY Address (The RX[2-3]-Lines are Floating on Reset)
		CheckPHYAddr();
		Seriell_Text("\r\nPHY_ADDR:");
		Seriell_IntToHex(phyAddr);

		/* Put the DP83848C in reset mode */
		write_PHY (PHY_REG_BMCR, 0x8000);
		//write_PHY (PHY_REG_BMCR, 0x8000);

		/* Wait for hardware reset to end. */
		iStep = 3;
	case 3:
		regv = read_PHY (PHY_REG_BMCR);
		if ((regv & 0x8000))
			return 0;

		/* Check if this is a DP83848C PHY. */
		id1 = read_PHY (PHY_REG_IDR1);
		id2 = read_PHY (PHY_REG_IDR2);
		DWORD id3 = (((DWORD)id1 << 16) | (id2 & 0xFFF0));

		phyTyp = 0;
		if ((id3 & DP83848C_ID) == DP83848C_ID)
			phyTyp = 1;
		else if ((id3 & 0x00FF1610) == KS8721_ID)
			phyTyp = 2;
		if (!phyTyp)
		{
			Seriell_Text("PHY-ID:");
			Seriell_LongToHex(id3);
			return 0;
		}
		/* Configure the PHY device */

		/* Use autonegotiation about the link speed. */
		write_PHY (PHY_REG_BMCR, PHY_AUTO_NEG);
		/* Wait to complete Auto_Negotiation. */
		iStep = 4;
	case 4:
		regv = read_PHY (PHY_REG_BMSR);
		if (!(regv & 0x0020))
		{
			/* Autonegotiation Complete. */
			return 0;
		}

		/* Check the link status. */
		iStep = 5;
	case 5:
		if (phyTyp == 1)
		{
			regv = read_PHY (PHY_REG_STS);
			if ((regv & 0x0011) != 0x0011)
				return 0;
			/* Link established from here on */
			if ( regv & 0x04 )
				Duplex = FULL_DUPLEX;
			else Duplex = HALF_DUPLEX;

			if ( regv & 0x02 )
				Speed = SPEED_10;
			else Speed = SPEED_100;

			/*// Configure Full/Half Duplex mode.
			if (regv & 0x0004) {
				// Full duplex is enabled.
				MAC_MAC2    |= MAC2_FULL_DUP;
				MAC_COMMAND |= CR_FULL_DUP;
				MAC_IPGT     = IPGT_FULL_DUP;
			} else {
				// Half duplex mode.
				MAC_IPGT = IPGT_HALF_DUP;
			}

			// Configure 100MBit/10MBit mode.
			if (regv & 0x0002) {
				// 10MBit mode.
				MAC_SUPP = 0;
			} else {
				// 100MBit mode.
				MAC_SUPP = SUPP_SPEED;
			}	*/
		} else if (phyTyp == 2)
		{
			regv = read_PHY (PHY_REG_BMSR);
			if (!((regv & 0x0024) == 0x0024))
				return 0;
			regv = read_PHY( MIC_PHY_100BASE_PHY_CTRL );
			/* successful negotiations; update link info */
			regv &= 0x001C;
			switch ( regv )
			{
			case 0x0004:
				Speed = SPEED_10;
				Duplex = HALF_DUPLEX;
				break;
			case 0x0008:
				Speed = SPEED_100;
				Duplex = HALF_DUPLEX;
				break;
			case 0x0014:
				Speed = SPEED_10;
				Duplex = FULL_DUPLEX;
				break;
			case 0x0018:
				Speed = SPEED_100;
				Duplex = FULL_DUPLEX;
				break;
			default:	// Should not come here, force to set default, 100 FULL_DUPLEX
				Speed = SPEED_100;
				Duplex = FULL_DUPLEX;
				break;
			}
		}
		iStep = 6;
	case 6:
		if ( (Speed == SPEED_10) && (Duplex == HALF_DUPLEX) )
		{
			MAC_MAC2 = 0x30;		/* half duplex, CRC and PAD enabled. */
			MAC_SUPP &= ~0x0100;	/* RMII Support Reg. speed is set to 10M */
#if RMII
			MAC_COMMAND |= 0x0240;
#else
			MAC_COMMAND |= 0x0040;	/* [10]-half duplex,[9]-MII mode,[6]-Pass runt frame, [5]-RxReset */
#endif
			/* back to back int-packet gap */
			MAC_IPGT = 0x0012;		/* IPG setting in half duplex mode */
		}
		else if ( (Speed == SPEED_100) && (Duplex == HALF_DUPLEX) )
		{
			MAC_MAC2 = 0x30;		/* half duplex, CRC and PAD enabled. */
			MAC_SUPP |= 0x0100;		/* RMII Support Reg. speed is set to 100M */
#if RMII
			MAC_COMMAND |= 0x0240;
#else
			MAC_COMMAND |= 0x0040;	/* [10]-half duplex,[9]-MII mode,[6]-Pass runt frame, [5]-RxReset */
#endif
			/* back to back int-packet gap */
			MAC_IPGT = 0x0012;		/* IPG setting in half duplex mode */
		}
		else if ( (Speed == SPEED_10) && (Duplex == FULL_DUPLEX) )
		{
			MAC_MAC2 = 0x31;		/* full duplex, CRC and PAD enabled. */
			MAC_SUPP &= ~0x0100;	/* RMII Support Reg. speed is set to 10M */
#if RMII
			MAC_COMMAND |= 0x0640;
#else
			MAC_COMMAND |= 0x0440;	/* [10]-full duplex,[9]-MII mode,[6]-Pass runt frame, [5]-RxReset */
#endif
			/* back to back int-packet gap */
			MAC_IPGT = 0x0015;		/* IPG setting in full duplex mode */
		}
		else if ( (Speed == SPEED_100) && (Duplex == FULL_DUPLEX) )
							/* default setting, 100 BASE, FULL DUPLEX */
		{
			MAC_MAC2 = 0x31;		/* full duplex, CRC and PAD enabled. */
			MAC_SUPP |= 0x0100;		/* RMII Support Reg. speed is set to 100M */
#if RMII
			MAC_COMMAND |= 0x0640;
#else
			MAC_COMMAND |= 0x0440;	/* [10]-full duplex,[9]-MII mode,[6]-Pass runt frame, [5]-RxReset */
#endif
			/* back to back int-packet gap */
			MAC_IPGT = 0x0015;	/* IPG setting in full duplex mode */
		}

		/* Set the Ethernet MAC Address registers */
		MAC_SA0 = (emacETHADDR0 << 8) | emacETHADDR1;
		MAC_SA1 = (emacETHADDR2 << 8) | emacETHADDR3;
		MAC_SA2 = (emacETHADDR4 << 8) | emacETHADDR5;

		/* Initialize Tx and Rx DMA Descriptors */
		rx_descr_init ();
		tx_descr_init ();

		/* Receive Broadcast and Perfect Match Packets */
		MAC_RXFILTERCTRL = RFC_UCAST_EN | RFC_BCAST_EN | RFC_PERFECT_EN;

		/* Create the semaphore used ot wake the uIP task. */

		/* Reset all interrupts */
		MAC_INTCLEAR  = 0xFFFF;

		/* Enable receive and transmit mode of MAC Ethernet core */
		MAC_COMMAND  |= (CR_RX_EN | CR_TX_EN);
		MAC_MAC1     |= MAC1_REC_EN;
		break;
	default:
		iStep = 0;
		return 0;
	}
	return 1;
}


// reads a word in little-endian byte order from RX_BUFFER

unsigned short ReadFrame_EMAC(void)
{
  return (*rptr++);
}

// reads a word in big-endian byte order from RX_FRAME_PORT
// (useful to avoid permanent byte-swapping while reading
// TCP/IP-data)

unsigned short ReadFrameBE_EMAC(void)
{
  unsigned short ReturnValue;

  ReturnValue = SwapBytes (*rptr++);
  return (ReturnValue);
}


// copies bytes from frame port to MCU-memory
// NOTES: * an odd number of byte may only be transfered
//          if the frame is read to the end!
//        * MCU-memory MUST start at word-boundary

void CopyFromFrame_EMAC(void *Dest, unsigned short Size)
{
  unsigned short * piDest;                       // Keil: Pointer added to correct expression

  piDest = Dest;                                 // Keil: Line added
  while (Size > 1) {
    *piDest++ = ReadFrame_EMAC();
    Size -= 2;
  }

  if (Size) {                                         // check for leftover byte...
    *(unsigned char *)piDest = (char)ReadFrame_EMAC();// the LAN-Controller will return 0
  }                                                   // for the highbyte
}

// does a dummy read on frame-I/O-port
// NOTE: only an even number of bytes is read!

void DummyReadFrame_EMAC(unsigned short Size)    // discards an EVEN number of bytes
{                                                // from RX-fifo
  while (Size > 1) {
    ReadFrame_EMAC();
    Size -= 2;
  }
}

// Reads the length of the received ethernet frame and checks if the
// destination address is a broadcast message or not
// returns the frame length
unsigned short StartReadFrame(void) {
  unsigned short RxLen;
  unsigned int idx;

  idx = MAC_RXCONSUMEINDEX;
  RxLen = (RX_STAT_INFO(idx) & RINFO_SIZE) + 1; //- 3;
  rptr = (unsigned short *)RX_DESC_PACKET(idx);
  return(RxLen);
}

void EndReadFrame(void) {
  unsigned int idx;

  /* DMA free packet. */
  idx = MAC_RXCONSUMEINDEX;

  if (++idx == NUM_RX_FRAG)
    idx = 0;

  MAC_RXCONSUMEINDEX = idx;
}

unsigned int CheckFrameReceived(void) {             // Packet received ?

  if (MAC_RXPRODUCEINDEX != MAC_RXCONSUMEINDEX)     // more packets received ?
    return(1);
  else
    return(0);
}

unsigned int uiGetEMACRxData( unsigned char *ucBuffer )
{
unsigned int uiLen = 0;

    if( MAC_RXPRODUCEINDEX != MAC_RXCONSUMEINDEX )
    {
        uiLen = StartReadFrame();
        CopyFromFrame_EMAC( ucBuffer, uiLen );
        EndReadFrame();
    }

    return uiLen;
}

// requests space in EMAC memory for storing an outgoing frame

void RequestSend(void)
{
  unsigned int idx;

  idx  = MAC_TXPRODUCEINDEX;
  tptr = (unsigned short *)TX_DESC_PACKET(idx);
}

// check if ethernet controller is ready to accept the
// frame we want to send

unsigned int Rdy4Tx(void)
{
  return (1);   // the ethernet controller transmits much faster
}               // than the CPU can load its buffers


// writes a word in little-endian byte order to TX_BUFFER
void WriteFrame_EMAC(unsigned short Data)
{
  *tptr++ = Data;
}

// copies bytes from MCU-memory to frame port
// NOTES: * an odd number of byte may only be transfered
//          if the frame is written to the end!
//        * MCU-memory MUST start at word-boundary

void CopyToFrame_EMAC(void *Source, unsigned int Size)
{
  unsigned short * piSource;

  piSource = Source;
  Size = (Size + 1) & 0xFFFE;    // round Size up to next even number
  while (Size > 0) {
    WriteFrame_EMAC(*piSource++);
    Size -= 2;
  }
}

void DoSend_EMAC(unsigned short FrameSize)
{
  unsigned int idx;

  idx = MAC_TXPRODUCEINDEX;
  TX_DESC_CTRL(idx) = FrameSize | TCTRL_LAST;
  if (++idx == NUM_TX_FRAG) idx = 0;
  MAC_TXPRODUCEINDEX = idx;
}

#endif
