/*
	FreeRTOS.org V5.0.2 - Copyright (C) 2003-2008 Richard Barry.

	This file is part of the FreeRTOS.org distribution.

	FreeRTOS.org is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	FreeRTOS.org is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with FreeRTOS.org; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	A special exception to the GPL can be applied should you wish to distribute
	a combined work that includes FreeRTOS.org, without being obliged to provide
	the source code for any proprietary components.  See the licensing section
	of http://www.FreeRTOS.org for full details of how and when the exception
	can be applied.

	***************************************************************************
	See http://www.FreeRTOS.org for documentation, latest information, license
	and contact details.  Please ensure to read the configuration and relevant
	port sections of the online documentation.
	***************************************************************************
*/
/* Standard includes. */
#include "Include.h"

#ifdef WITH_IP

#include <string.h>

/* uip includes. */
#include "uip.h"
#include "uip_arp.h"
#include "webserver.h"
#include "timer.h"
#include "clock-arch.h"

/* Demo includes. */
#include "emac.h"

/*-----------------------------------------------------------*/

/* IP address configuration. */
#define uipIP_ADDR0		192
#define uipIP_ADDR1		168
#define uipIP_ADDR2		1
#define uipIP_ADDR3		10

/* How long to wait before attempting to connect the MAC again. */
#define uipINIT_WAIT    100

struct __packed arp_hdr {
  struct uip_eth_hdr ethhdr;
  u16_t hwtype;
  u16_t protocol;
  u8_t hwlen;
  u8_t protolen;
  u16_t opcode;
  struct uip_eth_addr shwaddr;
  u16_t sipaddr[2];
  struct uip_eth_addr dhwaddr;
  u16_t dipaddr[2];
};

/* Shortcut to the header within the Rx buffer. */
#define xHeader	((struct uip_eth_hdr *) (void *)&uip_buf[ 0 ])
#define BUFIP	((struct uip_tcpip_hdr *)(void*)&uip_buf[UIP_LLH_LEN])
#define BUFARP	((struct arp_hdr *)(void*)&uip_buf[0])
#define BUFICMP	((struct uip_icmpip_hdr *)(void*)&uip_buf[UIP_LLH_LEN])

/* Standard constant. */
#define uipTOTAL_FRAME_HEADER_SIZE	54

/*-----------------------------------------------------------*/

int iuIPStep;
struct timer periodic_timer, arp_timer;

static void prvENET_Send(void);
void vEMAC_ISR_Wrapper(void);
void Display_Ethernet(char * Text);

/*-----------------------------------------------------------*/

clock_time_t clock_time( void )
{
	return tSysTic;
}

void uIP_TaskInit(void)
{
	uip_ipaddr_t xIPAddr;
//	uip_ipaddr_t ipaddr;
	Seriell_Text("\r\nStart Init uIP...");
	iuIPStep = 0;

	uip_init();
	uip_arp_init();
	uip_ipaddr(xIPAddr, uipIP_ADDR0, uipIP_ADDR1, uipIP_ADDR2, uipIP_ADDR3 );
	uip_sethostaddr( xIPAddr );
	uip_ipaddr(xIPAddr, 255, 255, 0, 0);
	uip_setnetmask(xIPAddr);

/*	uip_init();
	uip_ipaddr(ipaddr, 192,168,1,10);
	uip_sethostaddr(ipaddr);
	uip_ipaddr(ipaddr, 192,168,1,1);
	uip_setdraddr(ipaddr);
	uip_ipaddr(ipaddr, 255,255,255,0);
	uip_setnetmask(ipaddr);		*/

	Seriell_Text("\r\nStart Init Port 80 Listen...");

	httpd_init();
//	uip_listen(HTONS(80));
//	uip_listen(HTONS(1000));

	Seriell_Text("\r\nEnd Init uIP & http.");
}

void uip_log(char *m);
void uip_log(char *m)
{
	Seriell_Text("\r\nuIP: ");
	Seriell_Text(m);
}

void uIP_Task(void)
{
	if (!iuIPStep) // Initialisieren EMAC im Task
	{
		if (tPuls100ms)
		{
			if (Init_EMAC() != FALSE)
			//if (EMACInit() != FALSE)
			{
				iuIPStep++;	// Netzwerk initialisiert und eingesteckt
//				struct uip_eth_addr xAddr;
				// Configure the MAC address in the uIP stack.

				//EMAC_RxEnable();
				//EMAC_TxEnable();

/*				uip_init();

				xAddr.addr[0] = EMAC_ADDR12 >> 8;	// Defined in EMAC.H
				xAddr.addr[1] = EMAC_ADDR12 & 0xFF;
				xAddr.addr[2] = EMAC_ADDR34 >> 8;
				xAddr.addr[3] = EMAC_ADDR34 & 0xFF;
				xAddr.addr[4] = EMAC_ADDR56 >> 8;
				xAddr.addr[5] = EMAC_ADDR56 & 0xFF;
				uip_setethaddr(xAddr);

				uip_ipaddr_t ipaddr;
				uip_ipaddr(ipaddr, uipIP_ADDR0, uipIP_ADDR1, uipIP_ADDR2, uipIP_ADDR3);
				uip_sethostaddr(ipaddr);*/
//				uip_ipaddr(ipaddr, 192,168,1,1);
//				uip_setdraddr(ipaddr);
//				uip_ipaddr(ipaddr, 255,255,255,0);
//				uip_setnetmask(ipaddr);

//				uip_arp_init();

//				uip_listen(HTONS(80));

				MAC_INTENABLE = INT_RX_DONE;
				install_irq( EMAC_INT, (void *)vEMAC_ISR_Wrapper, HIGHEST_PRIORITY );

				/* Configure the MAC address in the uIP stack. */
				struct uip_eth_addr xAddr;
				xAddr.addr[ 0 ] = emacETHADDR0;
				xAddr.addr[ 1 ] = emacETHADDR1;
				xAddr.addr[ 2 ] = emacETHADDR2;
				xAddr.addr[ 3 ] = emacETHADDR3;
				xAddr.addr[ 4 ] = emacETHADDR4;
				xAddr.addr[ 5 ] = emacETHADDR5;
				uip_setethaddr( xAddr );

				timer_set( &periodic_timer, 500 );
				timer_set( &arp_timer, 10000 );

				Seriell_Text("\r\nEMAC is Init.");
			}
		}
	} else if (CheckFrameReceived()) // neue Daten im Empfangsbuffer des PHY
	{	// Fertig initialisiert
		/*uip_len = EMACReceive( (void *)uip_buf );
		if (PacketReceived)
		{
			PacketReceived = 0;
			uip_len = ReceiveLength;
		} else uip_len = 0;*/

		uip_len = uiGetEMACRxData(uip_buf);

		Display_Ethernet("Receive");

		if( uip_len > 0 )
		{
			if( xHeader->type == htons( UIP_ETHTYPE_IP ) )
			{
				uip_arp_ipin();
				uip_input();

				/* If the above function invocation resulted in data that
				should be sent out on the network, the global variable
				uip_len is set to a value > 0. */
				if( uip_len > 0 )
				{
					uip_arp_out();
					Display_Ethernet("Send IP");
					prvENET_Send();
					//EMACSend((void*)uip_buf, uip_len);
				}
			} else if( xHeader->type == htons( UIP_ETHTYPE_ARP ) )
			{
				uip_arp_arpin();

				/* If the above function invocation resulted in data that
				should be sent out on the network, the global variable
				uip_len is set to a value > 0. */
				if( uip_len > 0 )
				{
					//uip_arp_out();
					Display_Ethernet("Send ARP");
					prvENET_Send();
					//EMACSend((void*)uip_buf, uip_len);
				}
			}
		} else {
			if( timer_expired( &periodic_timer ) ) // 500ms
			{
				int i;
				timer_reset( &periodic_timer );
				//Seriell_Text("\r\nT500");
				for( i = 0; i < UIP_CONNS; i++ )
				{
					uip_periodic( i );

					/* If the above function invocation resulted in data that
					should be sent out on the network, the global variable
					uip_len is set to a value > 0. */
					if( uip_len > 0 )
					{
						uip_arp_out();
						prvENET_Send();
						//EMACSend((void*)uip_buf, uip_len);
					}
				}
				/* Call the ARP timer function every 10 seconds. */
				if( timer_expired( &arp_timer ))
				{
					timer_reset( &arp_timer );
					//Seriell_Text("\r\nT10000");
					uip_arp_timer();
				}
			} else {
				/* We did not receive a packet, and there was no periodic
				processing to perform.  Block for a fixed period.  If a packet
				is received during this period we will be woken by the ISR
				giving us the Semaphore. */
				//xSemaphoreTake( xEMACSemaphore, configTICK_RATE_HZ / 2 );
			}
		}
	}
//  	if (FIO0PIN & 0x01000000) // PHY Int Pin
//  		FIO3SET = 0x04000000; // LED GN ein Int
//	else FIO3CLR = 0x04000000;
  	if (0 && tPuls1s)
  	{
  		LCD_CLS(2);
  		LCD_Char('P');
  		LCD_IntToHex(read_PHY(0x01)); // Status-Register
  		LCD_IntToHex(read_PHY(0x1B)); // ISR-Register
  		LCD_IntToHex(read_PHY(0x00)); // Basic-Control

  		//LCD_IntToHex(ReadPHY(0x01)); // Status-Register
  		//LCD_IntToHex(ReadPHY(0x1B)); // ISR-Register
  		//LCD_IntToHex(ReadPHY(0x00)); // Basic-Control
  	}
  	if (0 && tPuls1s)
  	{
  		LCD_CLS(2);
  		LCD_Char('L');
  		//LCD_NibbleHex(MAC_STATUS & 0xF);
  		LCD_LongToHex(MAC_RSV); // Status-Register
  		LCD_ByteToHex(MAC_RXPRODUCEINDEX & 0xFF);
  		LCD_ByteToHex(MAC_RXCONSUMEINDEX & 0xFF);
/*  		Seriell_CRLF();
  		Seriell_Char('L');
  		Seriell_LongToHex(MAC_RSV); // Status-Register
  		Seriell_IntToHex(MAC_RXCONSUMEINDEX);
  		Seriell_IntToHex(MAC_RXPRODUCEINDEX);
  		Seriell_IntToHex(MAC_MAC1);
  		Seriell_IntToHex(MAC_MAC2);		*/
  	}
}

/*-----------------------------------------------------------*/
static void prvENET_Send(void)
{
    RequestSend();

    /* Copy the header into the Tx buffer. */
    CopyToFrame_EMAC( uip_buf, uipTOTAL_FRAME_HEADER_SIZE );
    if( uip_len > uipTOTAL_FRAME_HEADER_SIZE )
    {
        CopyToFrame_EMAC( uip_appdata, ( uip_len - uipTOTAL_FRAME_HEADER_SIZE ) );
    }

    DoSend_EMAC( uip_len );
}
/*-----------------------------------------------------------*/
//
void vApplicationProcessFormInput( char *pcInputString, long xInputLength )
{
char *c;

	// Process the form input sent by the IO page of the served HTML.

	c = strstr( pcInputString, "?" );
    if( c )
    {
    	// Turn LED's on or off in accordance with the check box status.
		if( strstr( c, "LED0=1" ) != NULL )
		{
			FIO3CLR = 0x02000000;
		}
		else
		{
			FIO3SET = 0x02000000;
		}

		if( strstr( c, "LED1=1" ) != NULL )
		{
			FIO3CLR = 0x04000000;
		}
		else
		{
			FIO3SET = 0x04000000;
		}

/*		if( strstr( c, "LED2=1" ) != NULL )
		{
			FIO3CLR = 0x06000000;
		}
		else
		{
			FIO3SET = 0x06000000;
		}	*/

		// Find the start of the text to be displayed on the LCD.
		if( (c = strstr( c, "LCD=" )) != NULL )
		{
			c += 4;
			LCD_CLS(2);
			// LCD_Text(c);
			while (*c) // Spezielle Ausgabe des Textes, da Umformatiert werden muss!
			{
				if (*c == '+') // + ist Space
					LCD_Char(' ');
				else if (*c == '%') // nach % folgt HEX-Code fr ASCII zeichen
				{					// Wird aber mit dem Zeichensatz des Displays dargestellt!!!
					c++;
					short b;
					if (*c > '9')
					{
						*c &= ~0x20;
						*c -= 'A';
						b = 10 + (short)*c;
					} else
						b = (short)*c - '0';
					b <<= 4;
					c++;
					if (*c > '9')
					{
						*c &= ~0x20;
						*c -= 'A';
						b += (10 + (short)*c);
					} else
						b += (short)*c - '0';
					LCD_Char(b);
				} else LCD_Char(*c);
				c++;
			}
		}

        // Terminate the file name for further processing within uIP.
        *c = 0x00;
    }
}
//*/
/*
void uIP_appcall(void)
{
	Seriell_Text("\r\nAddCall");
}
//*/

void vEMAC_ISR_Wrapper( void )
{
    /* Clear the interrupt. */
	//Seriell_Text("\r\nEMAC ISR:");
	//Seriell_IntToHex(MAC_INTSTATUS);
    MAC_INTCLEAR = 0xffff;
    //MAC_RXCONSUMEINDEX = MAC_RXPRODUCEINDEX;
    VICVectAddr = 0;
}

void Display_Ethernet(char * Text)
{
	return;
	if (uip_len > 0)
	{
		LCD_CLS(2);
		LCD_Text("EMAC: ");
		LCD_Int(uip_len);
		Seriell_CRLF();
		Seriell_Text(Text);
		Seriell_Text("\r\nEMAC: Len=");
		Seriell_Int(uip_len);
		Seriell_Text(" type=");
		Seriell_IntToHex(xHeader->type);
		Seriell_Text(" Source=");
		int i;
		for (i = 0; i < 6; i++)
			Seriell_ByteToHex(xHeader->src.addr[i]);
		Seriell_Text(" Dest=");
		for (i = 0; i < 6; i++)
			Seriell_ByteToHex(xHeader->dest.addr[i]);
		if (xHeader->type == htons(UIP_ETHTYPE_IP))
		{
			Seriell_Text("\r\nType IP ");
			if (BUFIP->proto == UIP_PROTO_ICMP)
				Seriell_Text("ICMP");
			else if (BUFIP->proto == UIP_PROTO_TCP)
				Seriell_Text("TCP");
			else if (BUFIP->proto == UIP_PROTO_UDP)
				Seriell_Text("UDP");
			else if (BUFIP->proto == UIP_PROTO_ICMP6)
				Seriell_Text("ICMP6");
			Seriell_Text(" Src:");
			Seriell_IP(BUFIP->srcipaddr);
			if (BUFIP->proto == UIP_PROTO_TCP)
			{
				Seriell_Text(" P:");
				Seriell_Int(htons(BUFIP->srcport));
			}
			Seriell_Text(" Dest:");
			Seriell_IP(BUFIP->destipaddr);
			if (BUFIP->proto == UIP_PROTO_TCP)
			{
				Seriell_Text(" P:");
				Seriell_Int(htons(BUFIP->destport));
			}
			if (BUFIP->proto == UIP_PROTO_ICMP)
			{
				Seriell_Text(" Typ: ");
				Seriell_Int(BUFICMP->type);
				Seriell_Text(" Code: ");
				Seriell_Int(BUFICMP->icode);
			}
		} else if (xHeader->type == htons(UIP_ETHTYPE_ARP))
		{
			Seriell_Text("\r\nType ARP Src:");
			Seriell_IP(BUFARP->sipaddr);
			Seriell_Text(" Dest:");
			Seriell_IP(BUFARP->dipaddr);
		}
		Seriell_BytesF(uip_buf, uip_len);
		Seriell_CRLF();
	}
}

#endif
